/** @jsxRuntime classic */
/** @jsx jsx */
import { graphql } from '@keystone-6/core'
import { jsx } from '@keystone-ui/core'
import {
  FieldContainer,
  FieldLabel,
  Select,
  TextInput,
  Checkbox,
  MultiSelect,
} from '@keystone-ui/fields'
import { type HTMLAttributes, type ReactElement, type ReactNode, useState } from 'react'
import { isValidURL } from '../isValidURL'

export type FormFieldValue =
  | string
  | number
  | boolean
  | null
  | readonly FormFieldValue[]
  | { [key: string]: FormFieldValue | undefined }

export type FormField<Value extends FormFieldValue, Options> = {
  kind: 'form'
  Input(props: {
    value: Value
    onChange(value: Value): void
    autoFocus: boolean
    /**
     * This will be true when validate has returned false and the user has attempted to close the form
     * or when the form is open and they attempt to save the item
     */
    forceValidation: boolean
  }): ReactElement | null
  /**
   * The options are config about the field that are available on the
   * preview props when rendering the toolbar and preview component
   */
  options: Options
  defaultValue: Value
  /**
   * validate will be called in two cases:
   * - on the client in the editor when a user is changing the value.
   *   Returning `false` will block closing the form
   *   and saving the item.
   * - on the server when a change is received before allowing it to be saved
   *   if `true` is returned
   * @param value The value of the form field. You should NOT trust
   * this value to be of the correct type because it could come from
   * a potentially malicious client
   */
  validate(value: unknown): boolean
}

export type FormFieldWithGraphQLField<Value extends FormFieldValue, Options> = FormField<
  Value,
  Options
> & {
  graphql: {
    output: graphql.Field<
      { value: Value },
      Record<string, graphql.Arg<graphql.InputType, boolean>>,
      graphql.OutputType,
      'value'
    >
    input: graphql.NullableInputType
  }
}

type InlineMarksConfig =
  | 'inherit'
  | {
      bold?: 'inherit'
      code?: 'inherit'
      italic?: 'inherit'
      strikethrough?: 'inherit'
      underline?: 'inherit'
      keyboard?: 'inherit'
      subscript?: 'inherit'
      superscript?: 'inherit'
    }

type BlockFormattingConfig = {
  alignment?: 'inherit'
  blockTypes?: 'inherit'
  headingLevels?: 'inherit' | (1 | 2 | 3 | 4 | 5 | 6)[]
  inlineMarks?: InlineMarksConfig
  listTypes?: 'inherit'
  softBreaks?: 'inherit'
}

export type ChildField = {
  kind: 'child'
  options:
    | {
        kind: 'block'
        placeholder: string
        formatting?: BlockFormattingConfig
        dividers?: 'inherit'
        links?: 'inherit'
        relationships?: 'inherit'
      }
    | {
        kind: 'inline'
        placeholder: string
        formatting?: {
          inlineMarks?: InlineMarksConfig
          softBreaks?: 'inherit'
        }
        links?: 'inherit'
        relationships?: 'inherit'
      }
}

export type ArrayField<ElementField extends ComponentSchema> = {
  kind: 'array'
  element: ElementField
  // this is written with unknown to avoid typescript being annoying about circularity or variance things
  itemLabel?(props: unknown): string
  label?: string
}

export type RelationshipField<Many extends boolean> = {
  kind: 'relationship'
  listKey: string
  selection: string | undefined
  label: string
  many: Many
}

export interface ObjectField<
  Fields extends Record<string, ComponentSchema> = Record<string, ComponentSchema>
> {
  kind: 'object'
  fields: Fields
}

export type ConditionalField<
  DiscriminantField extends FormField<string | boolean, any>,
  ConditionalValues extends {
    [Key in `${DiscriminantField['defaultValue']}`]: ComponentSchema;
  }
> = {
  kind: 'conditional'
  discriminant: DiscriminantField
  values: ConditionalValues
}

// this is written like this rather than ArrayField<ComponentSchema> to avoid TypeScript erroring about circularity
type ArrayFieldInComponentSchema = {
  kind: 'array'
  element: ComponentSchema
  // this is written with unknown to avoid typescript being annoying about circularity or variance things
  itemLabel?(props: unknown): string
  label?: string
}

export type ComponentSchema =
  | ChildField
  | FormField<any, any>
  | ObjectField
  | ConditionalField<FormField<any, any>, { [key: string]: ComponentSchema }>
  | RelationshipField<boolean>
  | ArrayFieldInComponentSchema

// this is written like this rather than ArrayField<ComponentSchemaForGraphQL> to avoid TypeScript erroring about circularity
type ArrayFieldInComponentSchemaForGraphQL = {
  kind: 'array'
  element: ComponentSchemaForGraphQL
  // this is written with unknown to avoid typescript being annoying about circularity or variance things
  itemLabel?(props: unknown): string
  label?: string
}

export type ComponentSchemaForGraphQL =
  | FormFieldWithGraphQLField<any, any>
  | ObjectField<Record<string, ComponentSchemaForGraphQL>>
  | ConditionalField<
      FormFieldWithGraphQLField<any, any>,
      { [key: string]: ComponentSchemaForGraphQL }
    >
  | RelationshipField<boolean>
  | ArrayFieldInComponentSchemaForGraphQL

export const fields = {
  text ({
    label,
    defaultValue = '',
  }: {
    label: string
    defaultValue?: string
  }): FormFieldWithGraphQLField<string, undefined> {
    return {
      kind: 'form',
      Input ({ value, onChange, autoFocus }) {
        return (
          <FieldContainer>
            <FieldLabel>{label}</FieldLabel>
            <TextInput
              autoFocus={autoFocus}
              value={value}
              onChange={event => {
                onChange(event.target.value)
              }}
            />
          </FieldContainer>
        )
      },
      options: undefined,
      defaultValue,
      validate (value) {
        return typeof value === 'string'
      },
      graphql: {
        input: graphql.String,
        output: graphql.field({ type: graphql.String }),
      },
    }
  },
  integer ({
    label,
    defaultValue = 0,
  }: {
    label: string
    defaultValue?: number
  }): FormFieldWithGraphQLField<number, undefined> {
    const validate = (value: unknown) => {
      return typeof value === 'number' && Number.isFinite(value)
    }
    return {
      kind: 'form',
      Input ({ value, onChange, autoFocus, forceValidation }) {
        const [blurred, setBlurred] = useState(false)
        const [inputValue, setInputValue] = useState(String(value))
        const showValidation = forceValidation || (blurred && !validate(value))

        return (
          <FieldContainer>
            <FieldLabel>{label}</FieldLabel>
            <TextInput
              onBlur={() => setBlurred(true)}
              autoFocus={autoFocus}
              value={inputValue}
              onChange={event => {
                const raw = event.target.value
                setInputValue(raw)
                if (/^[+-]?\d+$/.test(raw)) {
                  onChange(Number(raw))
                } else {
                  onChange(NaN)
                }
              }}
            />
            {showValidation && <span css={{ color: 'red' }}>Please specify an integer</span>}
          </FieldContainer>
        )
      },
      options: undefined,
      defaultValue,
      validate,
      graphql: {
        input: graphql.Int,
        output: graphql.field({ type: graphql.Int }),
      },
    }
  },
  url ({
    label,
    defaultValue = '',
  }: {
    label: string
    defaultValue?: string
  }): FormFieldWithGraphQLField<string, undefined> {
    const validate = (value: unknown) => {
      return typeof value === 'string' && (value === '' || isValidURL(value))
    }
    return {
      kind: 'form',
      Input ({ value, onChange, autoFocus, forceValidation }) {
        const [blurred, setBlurred] = useState(false)
        const showValidation = forceValidation || (blurred && !validate(value))
        return (
          <FieldContainer>
            <FieldLabel>{label}</FieldLabel>
            <TextInput
              onBlur={() => setBlurred(true)}
              autoFocus={autoFocus}
              value={value}
              onChange={event => {
                onChange(event.target.value)
              }}
            />
            {showValidation && <span css={{ color: 'red' }}>Please provide a valid URL</span>}
          </FieldContainer>
        )
      },
      options: undefined,
      defaultValue,
      validate,
      graphql: {
        input: graphql.String,
        output: graphql.field({ type: graphql.String }),
      },
    }
  },
  select<Option extends { label: string, value: string }> ({
    label,
    options,
    defaultValue,
  }: {
    label: string
    options: readonly Option[]
    defaultValue: Option['value']
  }): FormFieldWithGraphQLField<Option['value'], readonly Option[]> {
    const optionValuesSet = new Set(options.map(x => x.value))
    if (!optionValuesSet.has(defaultValue)) {
      throw new Error(
        `A defaultValue of ${defaultValue} was provided to a select field but it does not match the value of one of the options provided`
      )
    }
    return {
      kind: 'form',
      Input ({ value, onChange, autoFocus }) {
        return (
          <FieldContainer>
            <FieldLabel>{label}</FieldLabel>
            <Select
              autoFocus={autoFocus}
              value={options.find(option => option.value === value) || null}
              onChange={option => {
                if (option) {
                  onChange(option.value)
                }
              }}
              options={options}
            />
          </FieldContainer>
        )
      },
      options,
      defaultValue,
      validate (value) {
        return typeof value === 'string' && optionValuesSet.has(value)
      },
      graphql: {
        input: graphql.String,
        output: graphql.field({
          type: graphql.String,
          // TODO: investigate why this resolve is required here
          resolve ({ value }) {
            return value
          },
        }),
      },
    }
  },
  multiselect<Option extends { label: string, value: string }> ({
    label,
    options,
    defaultValue,
  }: {
    label: string
    options: readonly Option[]
    defaultValue: readonly Option['value'][]
  }): FormFieldWithGraphQLField<readonly Option['value'][], readonly Option[]> {
    const valuesToOption = new Map(options.map(x => [x.value, x]))
    return {
      kind: 'form',
      Input ({ value, onChange, autoFocus }) {
        return (
          <FieldContainer>
            <FieldLabel>{label}</FieldLabel>
            <MultiSelect
              autoFocus={autoFocus}
              value={value.map(x => valuesToOption.get(x)!)}
              options={options}
              onChange={options => {
                onChange(options.map(x => x.value))
              }}
            />
          </FieldContainer>
        )
      },
      options,
      defaultValue,
      validate (value) {
        return (
          Array.isArray(value) &&
          value.every(value => typeof value === 'string' && valuesToOption.has(value))
        )
      },
      graphql: {
        input: graphql.list(graphql.nonNull(graphql.String)),
        output: graphql.field({
          type: graphql.list(graphql.nonNull(graphql.String)),
          // TODO: investigate why this resolve is required here
          resolve ({ value }) {
            return value
          },
        }),
      },
    }
  },
  checkbox ({
    label,
    defaultValue = false,
  }: {
    label: string
    defaultValue?: boolean
  }): FormFieldWithGraphQLField<boolean, undefined> {
    return {
      kind: 'form',
      Input ({ value, onChange, autoFocus }) {
        return (
          <FieldContainer>
            <Checkbox
              checked={value}
              autoFocus={autoFocus}
              onChange={event => {
                onChange(event.target.checked)
              }}
            >
              {label}
            </Checkbox>
          </FieldContainer>
        )
      },
      options: undefined,
      defaultValue,
      validate (value) {
        return typeof value === 'boolean'
      },
      graphql: {
        input: graphql.Boolean,
        output: graphql.field({ type: graphql.Boolean }),
      },
    }
  },
  empty (): FormField<null, undefined> {
    return {
      kind: 'form',
      Input () {
        return null
      },
      options: undefined,
      defaultValue: null,
      validate (value) {
        return value === null || value === undefined
      },
    }
  },
  child (
    options:
      | {
          kind: 'block'
          placeholder: string
          formatting?: BlockFormattingConfig | 'inherit'
          dividers?: 'inherit'
          links?: 'inherit'
          relationships?: 'inherit'
        }
      | {
          kind: 'inline'
          placeholder: string
          formatting?:
            | 'inherit'
            | {
                inlineMarks?: InlineMarksConfig
                softBreaks?: 'inherit'
              }
          links?: 'inherit'
          relationships?: 'inherit'
        }
  ): ChildField {
    return {
      kind: 'child',
      options:
        options.kind === 'block'
          ? {
              kind: 'block',
              placeholder: options.placeholder,
              dividers: options.dividers,
              formatting:
                options.formatting === 'inherit'
                  ? {
                      blockTypes: 'inherit',
                      headingLevels: 'inherit',
                      inlineMarks: 'inherit',
                      listTypes: 'inherit',
                      alignment: 'inherit',
                      softBreaks: 'inherit',
                    }
                  : options.formatting,
              links: options.links,
              relationships: options.relationships,
            }
          : {
              kind: 'inline',
              placeholder: options.placeholder,
              formatting:
                options.formatting === 'inherit'
                  ? { inlineMarks: 'inherit', softBreaks: 'inherit' }
                  : options.formatting,
              links: options.links,
              relationships: options.relationships,
            },
    }
  },
  object<Fields extends Record<string, ComponentSchema>> (fields: Fields): ObjectField<Fields> {
    return { kind: 'object', fields }
  },
  conditional<
    DiscriminantField extends FormField<string | boolean, any>,
    ConditionalValues extends {
      [Key in `${DiscriminantField['defaultValue']}`]: ComponentSchema;
    }
  > (
    discriminant: DiscriminantField,
    values: ConditionalValues
  ): ConditionalField<DiscriminantField, ConditionalValues> {
    if (
      (discriminant.validate('true') || discriminant.validate('false')) &&
      (discriminant.validate(true) || discriminant.validate(false))
    ) {
      throw new Error(
        'The discriminant of a conditional field only supports string values, or boolean values, not both.'
      )
    }
    return {
      kind: 'conditional',
      discriminant,
      values: values,
    }
  },
  relationship<Many extends boolean | undefined = false> ({
    listKey,
    selection,
    label,
    many,
  }: {
    listKey: string
    label: string
    selection?: string
  } & (Many extends undefined | false ? { many?: Many } : { many: Many })): RelationshipField<
    Many extends true ? true : false
  > {
    return {
      kind: 'relationship',
      listKey,
      selection,
      label,
      many: (many ? true : false) as any,
    }
  },
  array<ElementField extends ComponentSchema> (
    element: ElementField,
    opts?: {
      itemLabel?: (props: GenericPreviewProps<ElementField, unknown>) => string
      label?: string
    }
  ): ArrayField<ElementField> {
    return { kind: 'array', element, itemLabel: opts?.itemLabel, label: opts?.label }
  },
}

export type ComponentBlock<
  Fields extends Record<string, ComponentSchema> = Record<string, ComponentSchema>
> = {
  preview: (props: any) => ReactElement | null
  schema: Fields
  label: string
} & (
  | {
      chromeless: true
      toolbar?: (props: { props: Record<string, any>, onRemove(): void }) => ReactElement
    }
  | {
      chromeless?: false
      toolbar?: (props: {
        props: Record<string, any>
        onShowEditMode(): void
        onRemove(): void
        isValid: boolean
      }) => ReactElement
    }
)

type ChildFieldPreviewProps<Schema extends ChildField, ChildFieldElement> = {
  readonly element: ChildFieldElement
  readonly schema: Schema
}

type FormFieldPreviewProps<Schema extends FormField<any, any>> = {
  readonly value: Schema['defaultValue']
  onChange(value: Schema['defaultValue']): void
  readonly options: Schema['options']
  readonly schema: Schema
}

type ObjectFieldPreviewProps<Schema extends ObjectField<any>, ChildFieldElement> = {
  readonly fields: {
    readonly [Key in keyof Schema['fields']]: GenericPreviewProps<
      Schema['fields'][Key],
      ChildFieldElement
    >;
  }
  onChange(value: {
    readonly [Key in keyof Schema['fields']]?: InitialOrUpdateValueFromComponentPropField<
      Schema['fields'][Key]
    >;
  }): void
  readonly schema: Schema
}

type ConditionalFieldPreviewProps<
  Schema extends ConditionalField<FormField<string | boolean, any>, any>,
  ChildFieldElement
> = {
  readonly [Key in keyof Schema['values']]: {
    readonly discriminant: DiscriminantStringToDiscriminantValue<Schema['discriminant'], Key>
    onChange<Discriminant extends Schema['discriminant']['defaultValue']>(
      discriminant: Discriminant,
      value?: InitialOrUpdateValueFromComponentPropField<Schema['values'][`${Discriminant}`]>
    ): void
    readonly options: Schema['discriminant']['options']
    readonly value: GenericPreviewProps<Schema['values'][Key], ChildFieldElement>
    readonly schema: Schema
  };
}[keyof Schema['values']]

// this is a separate type so that this is distributive
type RelationshipDataType<Many extends boolean> = Many extends true
  ? readonly HydratedRelationshipData[]
  : HydratedRelationshipData | null

type RelationshipFieldPreviewProps<Schema extends RelationshipField<boolean>> = {
  readonly value: RelationshipDataType<Schema['many']>
  onChange(relationshipData: RelationshipDataType<Schema['many']>): void
  readonly schema: Schema
}

type ArrayFieldPreviewProps<Schema extends ArrayField<ComponentSchema>, ChildFieldElement> = {
  readonly elements: readonly (GenericPreviewProps<Schema['element'], ChildFieldElement> & {
    readonly key: string
  })[]
  readonly onChange: (
    value: readonly {
      key: string | undefined
      value?: InitialOrUpdateValueFromComponentPropField<Schema['element']>
    }[]
  ) => void
  readonly schema: Schema
}

export type GenericPreviewProps<
  Schema extends ComponentSchema,
  ChildFieldElement
> = Schema extends ChildField
  ? ChildFieldPreviewProps<Schema, ChildFieldElement>
  : Schema extends FormField<infer Value, infer Options>
  ? FormFieldPreviewProps<Schema>
  : Schema extends ObjectField<infer Value>
  ? ObjectFieldPreviewProps<Schema, ChildFieldElement>
  : Schema extends ConditionalField<infer DiscriminantField, infer Values>
  ? ConditionalFieldPreviewProps<Schema, ChildFieldElement>
  : Schema extends RelationshipField<infer Many>
  ? RelationshipFieldPreviewProps<Schema>
  : Schema extends ArrayField<infer ElementField>
  ? ArrayFieldPreviewProps<Schema, ChildFieldElement>
  : never

export type PreviewProps<Schema extends ComponentSchema> = GenericPreviewProps<Schema, ReactNode>

export type InitialOrUpdateValueFromComponentPropField<Schema extends ComponentSchema> =
  Schema extends ChildField
    ? undefined
    : Schema extends FormField<infer Value, any>
    ? Value | undefined
    : Schema extends ObjectField<infer Value>
    ? {
        readonly [Key in keyof Value]?: InitialOrUpdateValueFromComponentPropField<Value[Key]>;
      }
    : Schema extends ConditionalField<infer DiscriminantField, infer Values>
    ? {
        readonly [Key in keyof Values]: {
          readonly discriminant: DiscriminantStringToDiscriminantValue<DiscriminantField, Key>
          readonly value?: InitialOrUpdateValueFromComponentPropField<Values[Key]>
        };
      }[keyof Values]
    : Schema extends RelationshipField<infer Many>
    ? Many extends true
      ? readonly HydratedRelationshipData[]
      : HydratedRelationshipData | null
    : Schema extends ArrayField<infer ElementField>
    ? readonly {
        key: string | undefined
        value?: InitialOrUpdateValueFromComponentPropField<ElementField>
      }[]
    : never

type DiscriminantStringToDiscriminantValue<
  DiscriminantField extends FormField<any, any>,
  DiscriminantString extends PropertyKey
> = DiscriminantField['defaultValue'] extends boolean
  ? 'true' extends DiscriminantString
    ? true
    : 'false' extends DiscriminantString
    ? false
    : never
  : DiscriminantString

export type PreviewPropsForToolbar<Schema extends ComponentSchema> = GenericPreviewProps<
  Schema,
  undefined
>

export type HydratedRelationshipData = {
  id: string
  label: string
  data: Record<string, any>
}

export type RelationshipData = {
  id: string
  label: string | undefined
  data: Record<string, any> | undefined
}

export function component<
  Schema extends {
    [Key in any]: ComponentSchema;
  }
> (
  options: {
    /** The preview component shown in the editor */
    preview: (props: PreviewProps<ObjectField<Schema>>) => ReactElement | null
    /** The schema for the props that the preview component, toolbar and rendered component will receive */
    schema: Schema
    /** The label to show in the insert menu and chrome around the block if chromeless is false */
    label: string
  } & (
    | {
        chromeless: true
        toolbar?: (props: {
          props: PreviewPropsForToolbar<ObjectField<Schema>>
          onRemove(): void
        }) => ReactElement
      }
    | {
        chromeless?: false
        toolbar?: (props: {
          props: PreviewPropsForToolbar<ObjectField<Schema>>
          onShowEditMode(): void
          onRemove(): void
        }) => ReactElement
      }
  )
): ComponentBlock<Schema> {
  return options as any
}

export const NotEditable = ({ children, ...props }: HTMLAttributes<HTMLDivElement>) => (
  <span css={{ userSelect: 'none' }} contentEditable={false} {...props}>
    {children}
  </span>
)

type Comp<Props> = (props: Props) => ReactElement | null

type ValueForRenderingFromComponentPropField<Schema extends ComponentSchema> =
  Schema extends ChildField
    ? ReactNode
    : Schema extends FormField<infer Value, any>
    ? Value
    : Schema extends ObjectField<infer Value>
    ? { readonly [Key in keyof Value]: ValueForRenderingFromComponentPropField<Value[Key]> }
    : Schema extends ConditionalField<infer DiscriminantField, infer Values>
    ? {
        readonly [Key in keyof Values]: {
          readonly discriminant: DiscriminantStringToDiscriminantValue<DiscriminantField, Key>
          readonly value: ValueForRenderingFromComponentPropField<Values[Key]>
        };
      }[keyof Values]
    : Schema extends RelationshipField<infer Many>
    ? Many extends true
      ? readonly HydratedRelationshipData[]
      : HydratedRelationshipData | null
    : Schema extends ArrayField<infer ElementField>
    ? readonly ValueForRenderingFromComponentPropField<ElementField>[]
    : never

export type ValueForComponentSchema<Schema extends ComponentSchema> = Schema extends ChildField
  ? null
  : Schema extends FormField<infer Value, any>
  ? Value
  : Schema extends ObjectField<infer Value>
  ? { readonly [Key in keyof Value]: ValueForRenderingFromComponentPropField<Value[Key]> }
  : Schema extends ConditionalField<infer DiscriminantField, infer Values>
  ? {
      readonly [Key in keyof Values]: {
        readonly discriminant: DiscriminantStringToDiscriminantValue<DiscriminantField, Key>
        readonly value: ValueForRenderingFromComponentPropField<Values[Key]>
      };
    }[keyof Values]
  : Schema extends RelationshipField<infer Many>
  ? Many extends true
    ? readonly HydratedRelationshipData[]
    : HydratedRelationshipData | null
  : Schema extends ArrayField<infer ElementField>
  ? readonly ValueForRenderingFromComponentPropField<ElementField>[]
  : never

type ExtractPropsForPropsForRendering<Props extends Record<string, ComponentSchema>> = {
  readonly [Key in keyof Props]: ValueForRenderingFromComponentPropField<Props[Key]>;
}

export type InferRenderersForComponentBlocks<
  ComponentBlocks extends Record<string, ComponentBlock<any>>
> = {
  [Key in keyof ComponentBlocks]: Comp<
    ExtractPropsForPropsForRendering<ComponentBlocks[Key]['schema']>
  >;
}
